#include "SamplePlugin.h"
#include "ShaderSystemTexturedFog.h"

using namespace Ogre;
using namespace OgreBites;

const String Sample_ShaderSystemTexturedFog::FOG_DISTANCE_SLIDER = "FogDistance";
const String Sample_ShaderSystemTexturedFog::ACTIVATE_FOG_BUTTON = "ActivateFog";
const String Sample_ShaderSystemTexturedFog::FOG_BACKGROUND_SLIDER = "FogBackground";
const String Sample_ShaderSystemTexturedFog::ACTIVATE_SKY_BUTTON = "ActivateSkyBox";
