#include "SamplePlugin.h"
#include "ShaderSystemMultiLight.h"

using namespace Ogre;
using namespace OgreBites;

const String Sample_ShaderSystemMultiLight::DEBUG_MODE_CHECKBOX = "DebugMode";
const String Sample_ShaderSystemMultiLight::NUM_OF_LIGHTS_SLIDER = "NumOfLights";
const String Sample_ShaderSystemMultiLight::TWIRL_LIGHTS_CHECKBOX = "TwirlLights";
